"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MetricsCollector = void 0;
const json_1 = require("./services/json");
const toad_scheduler_1 = require("toad-scheduler");
const sum_1 = __importDefault(require("lodash/sum"));
const isEmpty_1 = __importDefault(require("lodash/isEmpty"));
const round_1 = __importDefault(require("lodash/round"));
const COMPLETED_EV = 'global:completed';
class MetricsCollector {
    constructor(queues, _config) {
        this._config = _config;
        this._completionCbs = [];
        this._processingTimeGauge = new Map();
        this._isActive = false;
        this._taskFn = () => __awaiter(this, void 0, void 0, function* () {
            const metrics = yield this._collect();
            yield this._persist(metrics);
        });
        this._scheduler = new toad_scheduler_1.ToadScheduler();
        this._queues = queues.filter((q) => !_config.blacklist.includes(q.name));
    }
    startCollecting() {
        this._maybeCreateSchedulerJob();
        this._scheduler.addSimpleIntervalJob(this._schedulerJob);
        this._attachCompletionCbs();
        this._isActive = true;
    }
    stopCollecting() {
        this._scheduler.stop();
        this._detachCompletionCbs();
        this._isActive = false;
    }
    extract(queue, start = 0, end = -1) {
        return __awaiter(this, void 0, void 0, function* () {
            const key = this._buildPersistKey(queue);
            const metrics = yield this._redisClient.lrange(key, start, end);
            return metrics.map(json_1.JsonService.maybeParse).filter(Boolean);
        });
    }
    clear(queue) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this._redisClient.del(this._buildPersistKey(queue));
        });
    }
    clearAll() {
        return __awaiter(this, void 0, void 0, function* () {
            const pipeline = this._redisClient.pipeline();
            this._queues.forEach((queue) => {
                pipeline.del(this._buildPersistKey(queue.id));
            });
            yield pipeline.exec();
        });
    }
    set queues(queues) {
        this._queues = queues.filter((q) => !this._config.blacklist.includes(q.name));
        const queuesSet = new Set(this._queues.map(({ id }) => id));
        this._processingTimeGauge.forEach((_, queueId) => {
            if (!queuesSet.has(queueId)) {
                this._processingTimeGauge.delete(queueId);
            }
        });
        if (this._isActive) {
            this._attachCompletionCbs();
        }
    }
    _maybeCreateSchedulerJob() {
        if (!this._schedulerJob) {
            const task = new toad_scheduler_1.AsyncTask('collect-metrics', this._taskFn);
            this._schedulerJob = new toad_scheduler_1.SimpleIntervalJob(this._config.collectInterval, task);
        }
    }
    _collect() {
        return __awaiter(this, void 0, void 0, function* () {
            const timestamp = Date.now();
            return yield Promise.all(this._queues.map((queue) => __awaiter(this, void 0, void 0, function* () {
                const processingTime = this._extractProcessingTime(queue.id);
                this._processingTimeGauge.set(queue.id, []);
                return Object.assign({ timestamp, queue: queue.id, counts: yield queue.getJobCounts() }, processingTime);
            })));
        });
    }
    _persist(metrics) {
        return __awaiter(this, void 0, void 0, function* () {
            const client = this._redisClient;
            const lpopPipeline = client.pipeline();
            yield Promise.all(metrics.map((metric) => __awaiter(this, void 0, void 0, function* () {
                const key = this._buildPersistKey(metric.queue);
                const listLen = yield client.rpush(key, JSON.stringify(metric));
                if (listLen > this._config.maxMetrics) {
                    lpopPipeline.lpop(key);
                }
            })));
            yield lpopPipeline.exec();
        });
    }
    _attachCompletionCbs() {
        this._detachCompletionCbs();
        for (const queue of this._queues) {
            const cb = this._onJobComplete.bind(this, queue);
            this._completionCbs.push(cb);
            queue.on(COMPLETED_EV, cb);
        }
    }
    _detachCompletionCbs() {
        if (this._completionCbs.length > 0) {
            for (const queue of this._queues) {
                for (const cb of this._completionCbs) {
                    queue.off(COMPLETED_EV, cb);
                }
            }
        }
    }
    _onJobComplete(queue, jobId) {
        return __awaiter(this, void 0, void 0, function* () {
            const job = yield queue.getJob(jobId);
            if (!(job === null || job === void 0 ? void 0 : job.finishedOn) || !job.processedOn) {
                return;
            }
            const dur = job.finishedOn - job.processedOn;
            const gauge = this._processingTimeGauge;
            const stats = gauge.get(queue.id);
            if (!stats) {
                gauge.set(queue.id, [dur]);
            }
            else {
                stats.push(dur);
            }
        });
    }
    _extractProcessingTime(queue) {
        const stats = this._processingTimeGauge.get(queue);
        if (isEmpty_1.default(stats))
            return {};
        return {
            processingTime: this._normalizeProcessingTime(sum_1.default(stats) / stats.length),
            processingTimeMin: this._normalizeProcessingTime(Math.min(...stats)),
            processingTimeMax: this._normalizeProcessingTime(Math.max(...stats)),
        };
    }
    _normalizeProcessingTime(time) {
        return round_1.default(time, 2);
    }
    _buildPersistKey(queue) {
        return this._config.redisPrefix + queue;
    }
    get _redisClient() {
        return this._queues[0].client;
    }
}
exports.MetricsCollector = MetricsCollector;
//# sourceMappingURL=metrics-collector.js.map